from typing import Any, Callable, Iterable

from cion.exceptions import ValidatorError

InnerValidator = Callable[[Any], Any]


def length(
    minimum: int, maximum: int, error_message: str = "Length must be between {minimum} and {maximum}"
) -> InnerValidator:
    """Validates the length of a string

    The type of the value to be validated is not enforced as :obj:`str`
    Used as a constraint on a field with :class:`.Schema`

    Args:
        minimum (int): The minimum length that the string can be
        maximum (int): The maximum length that the string can be
        error_message (str): The error message that is raised
            The error message is formatted like `error_message.format(minimum=minimum, maximum=maximum)`

    Returns:
        InnerValidator: The inner function that is called when validating schema
    """

    def inner(value: str):
        if len(value) < minimum or len(value) > maximum:
            raise ValidatorError(error_message.format(minimum=minimum, maximum=maximum))
        return value

    return inner


def one_of(values: Iterable[Any], error_message: str = "Value must be one of {values}") -> InnerValidator:
    """Checks if the value is in a list of values

    Used as a constraint on a field with :class:`.Schema`
    No assumptions are made about the type of the value

    Args:
        values: An iterable containing valid values
        error_message (str): The error message that is raised
            The `values` variable can be used, and it is the values joined together by a string

    Returns:
        InnerValidator: The inner function that is called when validating schema
    """

    def inner(value: Any):
        if value not in values:
            raise ValidatorError(error_message.format(values=", ".join(str(v) for v in values)))
        return value

    return inner



from . import exceptions, types, validators
from .exceptions import ValidationError, ValidatorError
from .schema import Schema

__all__ = ("Schema", "ValidatorError", "ValidationError", "exceptions", "types", "validators")

__version__ = "0.0.2"



from collections import defaultdict

from cion import types
from cion.exceptions import ValidationError, ValidatorError


class Schema:
    """Class that defines schema for data"""

    fields: dict[str, dict]  #: The fields to be validated
    ignore_extra: bool = False  #: Whether or not to ignore extra keys in the data

    def __init__(self, data: dict, ignore_extra: bool = False) -> None:
        for value in self.fields.values():
            value.setdefault("type", types.string())
            value.setdefault("constraints", [])

            value["constraints"].insert(0, value["type"])

        self.__data = data
        self.ignore_extra = ignore_extra

    def validate(self):
        """Validates the data

        Runs through all of the data and checks it against the fields

        Raises:
            ValidationError: When there is an error when validating
            ValueError: When ``self.ignore_extra`` is `False`, and there are extra values in the data
        """
        errors = defaultdict(list)
        valid_data = {}

        for key, value in self.__data.items():
            field = self.fields.get(key)
            if field is None and self.ignore_extra is False:
                raise ValueError(f"Unexpected field with name {key} found")
            errored = False
            for validator in field["constraints"]:
                try:
                    value = validator(value)
                except ValidatorError as error:
                    errored = True
                    errors[key].append(error.message)
            if errored is False:
                valid_data[key] = value

        for key, field in self.fields.items():
            value = valid_data.get(key)
            if value is None:
                if field.get("default") is not None:
                    valid_data[key] = field.get("default")
                elif field.get("required") is True:
                    errors[key].append(["This field is required"])

        try:
            valid_data = self.after_load(valid_data)
        except NotImplementedError:
            pass

        if bool(errors):  # check if empty basically, returns true if filled
            raise ValidationError(errors, valid_data)

        return valid_data

    def after_load(self, data: dict):
        """Function called after data is validated

        The default implementation is to raise NotImplementedError

        Args:
            data (dict): The validated data

        """
        raise NotImplementedError()



import re

from setuptools import find_packages, setup

with open("cion/__init__.py") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open("README.md") as f:
    readme = f.read()

setup(
    name="cion",
    description="A data validation library",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="meizuflux",
    project_urls={
        "Documentation": "https://cionpy.readthedocs.io",
        "Issue Tracker": "https://github.com/meizuflux/cion/issues",
        "Source Code": "https://github.com/meizuflux/cion",
    },
    license="MIT",
    url="https://github.com/meizuflux/cion",
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    version=version,
    install_requires=[],
    extras_require={
        "docs": ["sphinx", "furo", "sphinx-copybutton"],
        "lint": [
            "pyright",
        ],
        "build": ["setuptools", "wheel", "build"],
        "format": ["black", "isort"],
    },
    python_requires=">=3.10",
)



from cion.exceptions import ValidatorError


def string(cast: bool = True):
    """A string type

    Used when defining fields in :class:`.Schema`

    Args:
        cast (bool): Whether or not to try to cast the value to a :obj:`str`

    Returns:
        InnerValidator: The inner validator
    """

    def inner(value: str):
        if not isinstance(value, str):
            if cast:
                value = str(value)
            else:
                raise ValidatorError("Field must be a valid string")
        return value

    return inner


def integer(cast: bool = False):
    """An integer type

    Used when defining fields in :class:`.Schema`

    Args:
        cast (bool): Whether or not to try to cast the value to an :obj:`int`

    Returns:
        InnerValidator: The inner validator
    """

    def inner(value: int):
        if not isinstance(value, int):
            if cast:
                value = int(value)
            else:
                raise ValidatorError("Field must be a valid integer")
        return value

    return inner
